@@ -18,10 +18,9 @@
  *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
-#ifndef __BLUETOOTH_H__
-#define __BLUETOOTH_H__
 
 #include <drivers/atmodem/atutil.h>
+#include <ofono/dbus.h>
 
 /* AG supported features bitmap. Bluetooth HFP 1.5 spec page 77 */
 #define AG_FEATURE_3WAY 0x1
@@ -63,11 +62,14 @@ enum hfp_indicator {
 
 struct hfp_data {
 	GAtChat *chat;
+	char *handsfree_path;
+	DBusMessage *slc_msg;
 	unsigned int ag_features;
 	unsigned int ag_mpty_features;
 	unsigned int hf_features;
 	unsigned char cind_pos[HFP_INDICATOR_LAST];
 	unsigned int cind_val[HFP_INDICATOR_LAST];
+	gboolean agent_registered;
 };
 
 extern void hfp_netreg_init();
@@ -78,5 +80,3 @@ extern void hfp_call_volume_exit();
 
 extern void hfp_voicecall_init();
 extern void hfp_voicecall_exit();
-
-#endif
