@@ -70,31 +70,50 @@ struct sim_file_op {
 };
 
 struct ofono_sim {
+	/* Contents of the SIM file system, in rough initialization order */
 	char *iccid;
-	char *imsi;
-	enum ofono_sim_phase phase;
-	unsigned char mnc_length;
-	GSList *own_numbers;
-	GSList *new_numbers;
-	GSList *service_numbers;
-	gboolean sdn_ready;
-	enum ofono_sim_state state;
-	enum ofono_sim_password_type pin_type;
-	gboolean locked_pins[OFONO_SIM_PASSWORD_INVALID];
+
 	char **language_prefs;
-	GQueue *simop_q;
-	gint simop_source;
-	unsigned char efmsisdn_length;
-	unsigned char efmsisdn_records;
 	unsigned char *efli;
 	unsigned char efli_length;
+
+	enum ofono_sim_password_type pin_type;
+	gboolean locked_pins[OFONO_SIM_PASSWORD_SIM_PUK]; /* Number of PINs */
+
+	enum ofono_sim_phase phase;
+	unsigned char mnc_length;
 	enum ofono_sim_cphs_phase cphs_phase;
 	unsigned char cphs_service_table[2];
+	unsigned char *efust;
+	unsigned char efust_length;
+	unsigned char *efest;
+	unsigned char efest_length;
+	unsigned char *efsst;
+	unsigned char efsst_length;
+
+	char *imsi;
+
+	GSList *own_numbers;
+	GSList *new_numbers;
+	unsigned char efmsisdn_length;
+	unsigned char efmsisdn_records;
+
+	GSList *service_numbers;
+	gboolean sdn_ready;
+
+	unsigned char *efimg;
+	unsigned short efimg_length;
+
+	enum ofono_sim_state state;
 	struct ofono_watchlist *state_watches;
+
+	GQueue *simop_q;
+	gint simop_source;
+
+	DBusMessage *pending;
 	const struct ofono_sim_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
-	DBusMessage *pending;
 };
 
 struct msisdn_set_request {
@@ -172,6 +191,28 @@ static gboolean password_is_pin(enum ofono_sim_password_type type)
 	return FALSE;
 }
 
+static enum ofono_sim_password_type puk2pin(enum ofono_sim_password_type type)
+{
+	switch (type) {
+	case OFONO_SIM_PASSWORD_SIM_PUK:
+		return OFONO_SIM_PASSWORD_SIM_PIN;
+	case OFONO_SIM_PASSWORD_PHFSIM_PUK:
+		return OFONO_SIM_PASSWORD_PHFSIM_PIN;
+	case OFONO_SIM_PASSWORD_SIM_PUK2:
+		return OFONO_SIM_PASSWORD_SIM_PIN2;
+	case OFONO_SIM_PASSWORD_PHNET_PUK:
+		return OFONO_SIM_PASSWORD_PHNET_PUK;
+	case OFONO_SIM_PASSWORD_PHNETSUB_PUK:
+		return OFONO_SIM_PASSWORD_PHNETSUB_PIN;
+	case OFONO_SIM_PASSWORD_PHSP_PUK:
+		return OFONO_SIM_PASSWORD_PHSP_PIN;
+	case OFONO_SIM_PASSWORD_PHCORP_PUK:
+		return OFONO_SIM_PASSWORD_PHCORP_PIN;
+	default:
+		return OFONO_SIM_PASSWORD_INVALID;
+	}
+}
+
 static char **get_own_numbers(GSList *own_numbers)
 {
 	int nelem = 0;
@@ -200,13 +241,10 @@ static char **get_locked_pins(struct ofono_sim *sim)
 	int nelem = 0;
 	char **ret;
 
-	for (i = 0; i < OFONO_SIM_PASSWORD_INVALID; i++) {
+	for (i = 1; i < OFONO_SIM_PASSWORD_SIM_PUK; i++) {
 		if (sim->locked_pins[i] == FALSE)
 			continue;
 
-		if (password_is_pin(i) == FALSE)
-			continue;
-
 		nelem += 1;
 	}
 
@@ -214,13 +252,10 @@ static char **get_locked_pins(struct ofono_sim *sim)
 
 	nelem = 0;
 
-	for (i = 0; i < OFONO_SIM_PASSWORD_INVALID; i++) {
+	for (i = 1; i < OFONO_SIM_PASSWORD_SIM_PUK; i++) {
 		if (sim->locked_pins[i] == FALSE)
 			continue;
 
-		if (password_is_pin(i) == FALSE)
-			continue;
-
 		ret[nelem] = g_strdup(sim_passwd_name(i));
 		nelem += 1;
 	}
@@ -514,6 +549,7 @@ static void sim_locked_cb(struct ofono_sim *sim, gboolean locked)
 
 	type = sim_string_to_passwd(typestr);
 
+	/* This is used by lock/unlock pin, no puks allowed */
 	sim->locked_pins[type] = locked;
 	__ofono_dbus_pending_reply(&sim->pending, reply);
 
@@ -851,49 +887,6 @@ check:
 	sim->new_numbers = NULL;
 }
 
-static void sim_ad_read_cb(int ok, int length, int record,
-				const unsigned char *data,
-				int record_length, void *userdata)
-{
-	struct ofono_sim *sim = userdata;
-	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *path = __ofono_atom_get_path(sim->atom);
-	int new_mnc_length;
-	char mcc[OFONO_MAX_MCC_LENGTH + 1];
-	char mnc[OFONO_MAX_MNC_LENGTH + 1];
-	const char *str;
-
-	if (!ok)
-		return;
-
-	if (length < 4)
-		return;
-
-	new_mnc_length = data[3] & 0xf;
-
-	if (sim->mnc_length == new_mnc_length)
-		return;
-
-	sim->mnc_length = new_mnc_length;
-
-	strncpy(mcc, sim->imsi, OFONO_MAX_MCC_LENGTH);
-	mcc[OFONO_MAX_MCC_LENGTH] = '\0';
-	strncpy(mnc, sim->imsi + OFONO_MAX_MCC_LENGTH, sim->mnc_length);
-	mnc[sim->mnc_length] = '\0';
-
-	str = mcc;
-	ofono_dbus_signal_property_changed(conn, path,
-						OFONO_SIM_MANAGER_INTERFACE,
-						"MobileCountryCode",
-						DBUS_TYPE_STRING, &str);
-
-	str = mnc;
-	ofono_dbus_signal_property_changed(conn, path,
-						OFONO_SIM_MANAGER_INTERFACE,
-						"MobileNetworkCode",
-						DBUS_TYPE_STRING, &str);
-}
-
 static gint service_number_compare(gconstpointer a, gconstpointer b)
 {
 	const struct service_number *sdn = a;
@@ -980,7 +973,48 @@ static void sim_own_numbers_update(struct ofono_sim *sim)
 			sim_msisdn_read_cb, sim);
 }
 
-static void sim_ready(void *user, enum ofono_sim_state new_state)
+static void sim_efimg_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+	unsigned char *efimg;
+	int num_records;
+
+	if (!ok)
+		return;
+
+	num_records = length / record_length;
+
+	/*
+	 * EFimg descriptors are 9 bytes long.
+	 * Byte 1 of the record is the number of descriptors per record.
+	 */
+	if ((record_length < 10) ||
+			((record_length % 9 != 2) && (record_length % 9 != 1)))
+		return;
+
+	if (sim->efimg == NULL) {
+		sim->efimg = g_try_malloc0(num_records * 9);
+
+		if (sim->efimg == NULL)
+			return;
+
+		sim->efimg_length = num_records * 9;
+	}
+
+	/*
+	 * TBD - if we have more than one descriptor per record,
+	 * pick the nicest one.  For now we use the first one.
+	 */
+
+	/* copy descriptor into slot for this record */
+	efimg = &sim->efimg[(record - 1) * 9];
+
+	memcpy(efimg, &data[1], 9);
+}
+
+static void sim_ready(enum ofono_sim_state new_state, void *user)
 {
 	struct ofono_sim *sim = user;
 
@@ -989,33 +1023,10 @@ static void sim_ready(void *user, enum ofono_sim_state new_state)
 
 	sim_own_numbers_update(sim);
 
-	ofono_sim_read(sim, SIM_EFAD_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_ad_read_cb, sim);
 	ofono_sim_read(sim, SIM_EFSDN_FILEID, OFONO_SIM_FILE_STRUCTURE_FIXED,
 			sim_sdn_read_cb, sim);
-}
-
-static void sim_cphs_information_read_cb(int ok, int length, int record,
-				const unsigned char *data,
-				int record_length, void *userdata)
-{
-	struct ofono_sim *sim = userdata;
-
-	sim->cphs_phase = OFONO_SIM_CPHS_PHASE_NONE;
-
-	if (!ok || length < 3)
-		goto ready;
-
-	if (data[0] == 0x01)
-		sim->cphs_phase = OFONO_SIM_CPHS_PHASE_1G;
-	else if (data[0] >= 0x02)
-		sim->cphs_phase = OFONO_SIM_CPHS_PHASE_2G;
-
-	memcpy(sim->cphs_service_table, data + 1, 2);
-
-ready:
-	sim_set_ready(sim);
+	ofono_sim_read(sim, SIM_EFIMG_FILEID, OFONO_SIM_FILE_STRUCTURE_FIXED,
+			sim_efimg_read_cb, sim);
 }
 
 static void sim_imsi_cb(const struct ofono_error *error, const char *imsi,
@@ -1037,11 +1048,30 @@ static void sim_imsi_cb(const struct ofono_error *error, const char *imsi,
 						"SubscriberIdentity",
 						DBUS_TYPE_STRING, &sim->imsi);
 
-	/* Read CPHS-support bits, this is still part of the SIM
-	 * initialisation but no order is specified for it.  */
-	ofono_sim_read(sim, SIM_EF_CPHS_INFORMATION_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_cphs_information_read_cb, sim);
+	if (sim->mnc_length) {
+		char mcc[OFONO_MAX_MCC_LENGTH + 1];
+		char mnc[OFONO_MAX_MNC_LENGTH + 1];
+		const char *str;
+
+		strncpy(mcc, sim->imsi, OFONO_MAX_MCC_LENGTH);
+		mcc[OFONO_MAX_MCC_LENGTH] = '\0';
+		strncpy(mnc, sim->imsi + OFONO_MAX_MCC_LENGTH, sim->mnc_length);
+		mnc[sim->mnc_length] = '\0';
+
+		str = mcc;
+		ofono_dbus_signal_property_changed(conn, path,
+						OFONO_SIM_MANAGER_INTERFACE,
+						"MobileCountryCode",
+						DBUS_TYPE_STRING, &str);
+
+		str = mnc;
+		ofono_dbus_signal_property_changed(conn, path,
+						OFONO_SIM_MANAGER_INTERFACE,
+						"MobileNetworkCode",
+						DBUS_TYPE_STRING, &str);
+	}
+
+	sim_set_ready(sim);
 }
 
 static void sim_retrieve_imsi(struct ofono_sim *sim)
@@ -1055,6 +1085,157 @@ static void sim_retrieve_imsi(struct ofono_sim *sim)
 	sim->driver->read_imsi(sim, sim_imsi_cb, sim);
 }
 
+static void sim_efsst_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+
+	if (!ok)
+		goto out;
+
+	if (length < 2) {
+		ofono_error("EFsst shall contain at least two bytes");
+		goto out;
+	}
+
+	sim->efsst = g_memdup(data, length);
+	sim->efsst_length = length;
+
+out:
+	sim_retrieve_imsi(sim);
+}
+
+static void sim_efest_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+
+	if (!ok)
+		goto out;
+
+	if (length < 1) {
+		ofono_error("EFest shall contain at least one byte");
+		goto out;
+	}
+
+	sim->efest = g_memdup(data, length);
+	sim->efest_length = length;
+
+out:
+	sim_retrieve_imsi(sim);
+}
+
+static void sim_efust_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+
+	if (!ok)
+		goto out;
+
+	if (length < 1) {
+		ofono_error("EFust shall contain at least one byte");
+		goto out;
+	}
+
+	sim->efust = g_memdup(data, length);
+	sim->efust_length = length;
+
+	ofono_sim_read(sim, SIM_EFEST_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_efest_read_cb, sim);
+
+	return;
+
+out:
+	sim_retrieve_imsi(sim);
+}
+
+static void sim_cphs_information_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+
+	sim->cphs_phase = OFONO_SIM_CPHS_PHASE_NONE;
+
+	if (!ok || length < 3)
+		return;
+
+	if (data[0] == 0x01)
+		sim->cphs_phase = OFONO_SIM_CPHS_PHASE_1G;
+	else if (data[0] >= 0x02)
+		sim->cphs_phase = OFONO_SIM_CPHS_PHASE_2G;
+
+	memcpy(sim->cphs_service_table, data + 1, 2);
+}
+
+static void sim_ad_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+	int new_mnc_length;
+
+	if (!ok)
+		return;
+
+	if (length < 4)
+		return;
+
+	new_mnc_length = data[3] & 0xf;
+
+	/* sanity check for potential invalid values */
+	if (new_mnc_length < 2 || new_mnc_length > 3)
+		return;
+
+	sim->mnc_length = new_mnc_length;
+}
+
+static void sim_efphase_read_cb(int ok, int length, int record,
+				const unsigned char *data,
+				int record_length, void *userdata)
+{
+	struct ofono_sim *sim = userdata;
+
+	if (!ok || length != 1)
+		sim->phase = OFONO_SIM_PHASE_3G;
+	else
+		sim->phase = data[0];
+}
+
+static void sim_initialize_after_pin(struct ofono_sim *sim)
+{
+	ofono_sim_read(sim, SIM_EFPHASE_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_efphase_read_cb, sim);
+
+	ofono_sim_read(sim, SIM_EFAD_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_ad_read_cb, sim);
+
+	/*
+	 * Read CPHS-support bits, this is still part of the SIM
+	 * initialisation but no order is specified for it.
+	 */
+	ofono_sim_read(sim, SIM_EF_CPHS_INFORMATION_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_cphs_information_read_cb, sim);
+
+	/* Also retrieve the GSM service table */
+	if (sim->phase >= OFONO_SIM_PHASE_3G)
+		ofono_sim_read(sim, SIM_EFUST_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+				sim_efust_read_cb, sim);
+	else
+		ofono_sim_read(sim, SIM_EFSST_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+				sim_efsst_read_cb, sim);
+}
+
 static void sim_pin_query_cb(const struct ofono_error *error,
 				enum ofono_sim_password_type pin_type,
 				void *data)
@@ -1074,7 +1255,12 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 		sim->pin_type = pin_type;
 		pin_name = sim_passwd_name(pin_type);
 
-		sim->locked_pins[pin_type] = TRUE;
+		if (pin_type != OFONO_SIM_PASSWORD_NONE &&
+				password_is_pin(pin_type) == FALSE)
+			pin_type = puk2pin(pin_type);
+
+		if (pin_type != OFONO_SIM_PASSWORD_INVALID)
+			sim->locked_pins[pin_type] = TRUE;
 
 		ofono_dbus_signal_property_changed(conn, path,
 						OFONO_SIM_MANAGER_INTERFACE,
@@ -1084,13 +1270,13 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 
 checkdone:
 	if (pin_type == OFONO_SIM_PASSWORD_NONE)
-		sim_retrieve_imsi(sim);
+		sim_initialize_after_pin(sim);
 }
 
 static void sim_pin_check(struct ofono_sim *sim)
 {
 	if (!sim->driver->query_passwd_state) {
-		sim_retrieve_imsi(sim);
+		sim_initialize_after_pin(sim);
 		return;
 	}
 
@@ -1270,31 +1456,14 @@ skip_efpl:
 		g_slist_free(efpl);
 	}
 
-	if (sim->language_prefs == NULL)
-		return;
-
-	ofono_dbus_signal_array_property_changed(conn, path,
+	if (sim->language_prefs != NULL)
+		ofono_dbus_signal_array_property_changed(conn, path,
 						OFONO_SIM_MANAGER_INTERFACE,
 						"PreferredLanguages",
 						DBUS_TYPE_STRING,
 						&sim->language_prefs);
-}
 
-static void sim_retrieve_efli_and_efpl(struct ofono_sim *sim)
-{
-	/* According to 31.102 the EFli is read first and EFpl is then
-	 * only read if none of the EFli languages are supported by user
-	 * interface.  51.011 mandates the exact opposite, making EFpl/EFelp
-	 * preferred over EFlp (same EFid as EFli, different format).
-	 * However we don't depend on the user interface and so
-	 * need to read both files now.
-	 */
-	ofono_sim_read(sim, SIM_EFLI_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_efli_read_cb, sim);
-	ofono_sim_read(sim, SIM_EFPL_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_efpl_read_cb, sim);
+	sim_pin_check(sim);
 }
 
 static void sim_iccid_read_cb(int ok, int length, int record,
@@ -1320,36 +1489,6 @@ static void sim_iccid_read_cb(int ok, int length, int record,
 						&sim->iccid);
 }
 
-static void sim_efphase_read_cb(const struct ofono_error *error,
-				const unsigned char *data, int len, void *user)
-{
-	struct ofono_sim *sim = user;
-
-	if (!error || error->type != OFONO_ERROR_TYPE_NO_ERROR || len != 1)
-		sim->phase = OFONO_SIM_PHASE_3G;
-	else
-		sim->phase = data[0];
-
-	/* Proceed with SIM initialization */
-	ofono_sim_read(sim, SIM_EF_ICCID_FILEID,
-			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
-			sim_iccid_read_cb, sim);
-
-	sim_retrieve_efli_and_efpl(sim);
-	sim_pin_check(sim);
-}
-
-static void sim_determine_phase(struct ofono_sim *sim)
-{
-	if (!sim->driver->read_file_transparent) {
-		sim_efphase_read_cb(NULL, NULL, 0, sim);
-		return;
-	}
-
-	sim->driver->read_file_transparent(sim, SIM_EFPHASE_FILEID, 0, 1,
-						sim_efphase_read_cb, sim);
-}
-
 static void sim_initialize(struct ofono_sim *sim)
 {
 	/* Perform SIM initialization according to 3GPP 31.102 Section 5.1.1.2
@@ -1358,20 +1497,45 @@ static void sim_initialize(struct ofono_sim *sim)
 	 * responsible for checking the PIN, reading the IMSI and signaling
 	 * SIM ready condition.
 	 *
-	 * The procedure according to 31.102 is roughly:
+	 * The procedure according to 31.102, 51.011, 11.11 and CPHS 4.2 is
+	 * roughly:
+	 *
 	 * Read EFecc
 	 * Read EFli and EFpl
 	 * SIM Pin check
 	 * Request SIM phase (only in 51.011)
-	 * Read EFust
-	 * Read EFest
+	 * Administrative information request (read EFad)
+	 * Request CPHS Information (only in CPHS 4.2)
+	 * Read EFsst (only in 11.11 & 51.011)
+	 * Read EFust (only in 31.102)
+	 * Read EFest (only in 31.102)
 	 * Read IMSI
 	 *
 	 * At this point we signal the SIM ready condition and allow
 	 * arbitrary files to be written or read, assuming their presence
 	 * in the EFust
 	 */
-	sim_determine_phase(sim);
+
+	/* Grab the EFiccid which is always available */
+	ofono_sim_read(sim, SIM_EF_ICCID_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_iccid_read_cb, sim);
+
+	/* EFecc is read by the voicecall atom */
+
+	/* According to 31.102 the EFli is read first and EFpl is then
+	 * only read if none of the EFli languages are supported by user
+	 * interface.  51.011 mandates the exact opposite, making EFpl/EFelp
+	 * preferred over EFlp (same EFid as EFli, different format).
+	 * However we don't depend on the user interface and so
+	 * need to read both files now.
+	 */
+	ofono_sim_read(sim, SIM_EFLI_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_efli_read_cb, sim);
+	ofono_sim_read(sim, SIM_EFPL_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			sim_efpl_read_cb, sim);
 }
 
 static void sim_op_error(struct ofono_sim *sim)
@@ -1908,11 +2072,37 @@ static void sim_free_state(struct ofono_sim *sim)
 		g_strfreev(sim->language_prefs);
 		sim->language_prefs = NULL;
 	}
+
+	if (sim->efust) {
+		g_free(sim->efust);
+		sim->efust = NULL;
+		sim->efust_length = 0;
+	}
+
+	if (sim->efest) {
+		g_free(sim->efest);
+		sim->efest = NULL;
+		sim->efest_length = 0;
+	}
+
+	if (sim->efsst) {
+		g_free(sim->efsst);
+		sim->efsst = NULL;
+		sim->efsst_length = 0;
+	}
+
+	sim->mnc_length = 0;
+
+	if (sim->efimg) {
+		g_free(sim->efimg);
+		sim->efimg = NULL;
+		sim->efimg_length = 0;
+	}
 }
 
 void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 {
-	ofono_sim_state_event_notify_cb_t notify;
+	ofono_sim_state_event_cb_t notify;
 	GSList *l;
 
 	if (inserted == TRUE && sim->state == OFONO_SIM_STATE_NOT_PRESENT)
@@ -1931,7 +2121,7 @@ void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 		struct ofono_watchlist_item *item = l->data;
 		notify = item->notify;
 
-		notify(item->notify_data, sim->state);
+		notify(sim->state, item->notify_data);
 	}
 
 	if (inserted)
@@ -1941,8 +2131,8 @@ void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 }
 
 unsigned int ofono_sim_add_state_watch(struct ofono_sim *sim,
-				ofono_sim_state_event_notify_cb_t notify,
-				void *data, ofono_destroy_func destroy)
+					ofono_sim_state_event_cb_t notify,
+					void *data, ofono_destroy_func destroy)
 {
 	struct ofono_watchlist_item *item;
 
@@ -1979,7 +2169,7 @@ enum ofono_sim_state ofono_sim_get_state(struct ofono_sim *sim)
 static void sim_set_ready(struct ofono_sim *sim)
 {
 	GSList *l;
-	ofono_sim_state_event_notify_cb_t notify;
+	ofono_sim_state_event_cb_t notify;
 
 	if (sim == NULL)
 		return;
@@ -1993,7 +2183,7 @@ static void sim_set_ready(struct ofono_sim *sim)
 		struct ofono_watchlist_item *item = l->data;
 		notify = item->notify;
 
-		notify(item->notify_data, sim->state);
+		notify(sim->state, item->notify_data);
 	}
 }
 
