@@ -32,6 +32,7 @@
 #include <gdbus.h>
 #include <errno.h>
 #include <time.h>
+#include <sys/time.h>
 
 #include "ofono.h"
 
@@ -65,10 +66,12 @@ struct ofono_stk {
 	struct stk_agent *default_agent;
 	struct stk_agent *current_agent; /* Always equals one of the above */
 	struct stk_menu *main_menu, *select_item_menu;
+	gboolean respond_on_exit;
 	ofono_bool_t immediate_response;
 	guint remove_agent_source;
 	struct sms_submit_req *sms_submit_req;
 	char *idle_mode_text;
+	struct timeval get_inkey_start_ts;
 };
 
 struct envelope_op {
@@ -95,6 +98,8 @@ static int stk_respond(struct ofono_stk *stk, struct stk_response *rsp,
 	const guint8 *tlv;
 	unsigned int tlv_len;
 
+	DBG("");
+
 	if (stk->driver->terminal_response == NULL)
 		return -ENOSYS;
 
@@ -110,6 +115,7 @@ static int stk_respond(struct ofono_stk *stk, struct stk_response *rsp,
 
 	stk_command_free(stk->pending_cmd);
 	stk->pending_cmd = NULL;
+	stk->cancel_cmd = NULL;
 
 	stk->driver->terminal_response(stk, tlv_len, tlv, cb, stk);
 
@@ -132,6 +138,8 @@ static void send_simple_response(struct ofono_stk *stk,
 	struct stk_response rsp;
 	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
 
+	DBG("result %d", result);
+
 	memset(&rsp, 0, sizeof(rsp));
 	rsp.result.type = result;
 
@@ -146,6 +154,8 @@ static void envelope_cb(const struct ofono_error *error, const uint8_t *data,
 	struct envelope_op *op = g_queue_peek_head(stk->envelope_q);
 	gboolean result = TRUE;
 
+	DBG("length %d", length);
+
 	if (op->retries > 0 && error->type == OFONO_ERROR_TYPE_SIM &&
 			error->error == 0x9300) {
 		op->retries--;
@@ -185,6 +195,8 @@ static int stk_send_envelope(struct ofono_stk *stk, struct stk_envelope *e,
 	unsigned int tlv_len;
 	struct envelope_op *op;
 
+	DBG("");
+
 	if (stk->driver->envelope == NULL)
 		return -ENOSYS;
 
@@ -228,6 +240,8 @@ void __ofono_cbs_sim_download(struct ofono_stk *stk, const struct cbs *msg)
 	struct stk_envelope e;
 	int err;
 
+	DBG("");
+
 	memset(&e, 0, sizeof(e));
 
 	e.type = STK_ENVELOPE_TYPE_CBS_PP_DOWNLOAD;
@@ -249,6 +263,8 @@ static struct stk_menu *stk_menu_create(const char *title,
 	GSList *l;
 	int i;
 
+	DBG("");
+
 	ret->title = g_strdup(title ? title : "");
 	ret->icon_id = 0;
 	ret->items = g_new0(struct stk_menu_item, g_slist_length(items) + 1);
@@ -414,44 +430,32 @@ static void stk_request_cancel(struct ofono_stk *stk)
 		stk_agent_request_cancel(stk->default_agent);
 }
 
-static gboolean agent_called(struct ofono_stk *stk)
-{
-	if (stk->pending_cmd == NULL)
-		return FALSE;
-
-	switch (stk->pending_cmd->type) {
-	case STK_COMMAND_TYPE_SELECT_ITEM:
-	case STK_COMMAND_TYPE_DISPLAY_TEXT:
-		return TRUE;
-	}
-
-	return FALSE;
-}
-
 static void default_agent_notify(gpointer user_data)
 {
 	struct ofono_stk *stk = user_data;
 
-	if (stk->current_agent == stk->default_agent && agent_called(stk))
+	if (stk->current_agent == stk->default_agent && stk->respond_on_exit)
 		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
 
 	stk->default_agent = NULL;
 	stk->current_agent = stk->session_agent;
+	stk->respond_on_exit = FALSE;
 }
 
 static void session_agent_notify(gpointer user_data)
 {
 	struct ofono_stk *stk = user_data;
 
-	ofono_debug("Session Agent removed");
+	DBG("Session Agent removed");
 
-	if (stk->current_agent == stk->session_agent && agent_called(stk)) {
-		ofono_debug("Sending Terminate response for session agent");
+	if (stk->current_agent == stk->session_agent && stk->respond_on_exit) {
+		DBG("Sending Terminate response for session agent");
 		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
 	}
 
 	stk->session_agent = NULL;
 	stk->current_agent = stk->default_agent;
+	stk->respond_on_exit = FALSE;
 
 	if (stk->remove_agent_source) {
 		g_source_remove(stk->remove_agent_source);
@@ -540,6 +544,8 @@ static void menu_selection_envelope_cb(struct ofono_stk *stk, gboolean ok,
 	const char *agent_path;
 	DBusMessage *reply;
 
+	DBG("");
+
 	if (!ok) {
 		ofono_error("Sending Menu Selection to UICC failed");
 
@@ -588,11 +594,13 @@ static DBusMessage *stk_select_item(DBusConnection *conn,
 	struct stk_envelope e;
 	struct stk_menu *menu = stk->main_menu;
 
-	if (stk->pending)
+	DBG("");
+
+	if (stk->pending || stk->session_agent)
 		return __ofono_error_busy(msg);
 
-	if (stk->session_agent || !menu)
-		return __ofono_error_busy(msg);
+	if (!menu)
+		return __ofono_error_not_supported(msg);
 
 	if (dbus_message_get_args(msg, NULL,
 					DBUS_TYPE_BYTE, &selection,
@@ -614,6 +622,8 @@ static DBusMessage *stk_select_item(DBusConnection *conn,
 	e.menu_selection.item_id = menu->items[selection].item_id;
 	e.menu_selection.help_request = FALSE;
 
+	DBG("");
+
 	if (stk_send_envelope(stk, &e, menu_selection_envelope_cb, 0))
 		return __ofono_error_failed(msg);
 
@@ -665,14 +675,18 @@ static void send_sms_submit_cb(gboolean ok, void *data)
 	struct ofono_error failure = { .type = OFONO_ERROR_TYPE_FAILURE };
 	struct stk_response rsp;
 
-	ofono_debug("SMS submission %s", ok ? "successful" : "failed");
+	DBG("SMS submission %s", ok ? "successful" : "failed");
 
 	if (req->cancelled) {
-		ofono_debug("Received an SMS submitted callback after the "
+		DBG("Received an SMS submitted callback after the "
 				"proactive command was cancelled");
 		return;
 	}
 
+	if (stk->pending_cmd->send_sms.alpha_id &&
+			stk->pending_cmd->send_sms.alpha_id[0])
+		stk_alpha_id_unset(stk);
+
 	memset(&rsp, 0, sizeof(rsp));
 
 	if (ok == FALSE)
@@ -680,10 +694,6 @@ static void send_sms_submit_cb(gboolean ok, void *data)
 
 	if (stk_respond(stk, &rsp, stk_command_cb))
 		stk_command_cb(&failure, stk);
-
-	if (stk->pending_cmd->send_sms.alpha_id &&
-			stk->pending_cmd->send_sms.alpha_id[0])
-		stk_alpha_id_unset(stk);
 }
 
 static gboolean handle_command_send_sms(const struct stk_command *cmd,
@@ -972,6 +982,8 @@ static void request_selection_cb(enum stk_agent_result result, uint8_t id,
 {
 	struct ofono_stk *stk = user_data;
 
+	stk->respond_on_exit = FALSE;
+
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 	{
@@ -1017,18 +1029,21 @@ static gboolean handle_command_select_item(const struct stk_command *cmd,
 		return TRUE;
 	}
 
-	stk->cancel_cmd = stk_request_cancel;
-
 	/* We most likely got an out of memory error, tell SIM to retry */
 	if (stk_agent_request_selection(stk->current_agent,
 					stk->select_item_menu,
 					request_selection_cb, stk,
 					request_selection_destroy,
 					stk->timeout * 1000) < 0) {
+		request_selection_destroy(stk);
+
 		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
 		return TRUE;
 	}
 
+	stk->cancel_cmd = stk_request_cancel;
+	stk->respond_on_exit = TRUE;
+
 	return FALSE;
 }
 
@@ -1044,6 +1059,8 @@ static void display_text_cb(enum stk_agent_result result, void *user_data)
 	struct ofono_stk *stk = user_data;
 	gboolean confirm;
 
+	stk->respond_on_exit = FALSE;
+
 	/*
 	 * There are four possible paths for DisplayText with immediate
 	 * response flag set:
@@ -1116,8 +1133,6 @@ static gboolean handle_command_display_text(const struct stk_command *cmd,
 		}
 	}
 
-	stk->cancel_cmd = stk_request_cancel;
-
 	/* We most likely got an out of memory error, tell SIM to retry */
 	if (stk_agent_display_text(stk->current_agent, dt->text, 0, priority,
 					display_text_cb, stk,
@@ -1131,9 +1146,450 @@ static gboolean handle_command_display_text(const struct stk_command *cmd,
 
 	DBG("Immediate Response: %d", stk->immediate_response);
 
+	if (stk->immediate_response == FALSE) {
+		stk->respond_on_exit = TRUE;
+		stk->cancel_cmd = stk_request_cancel;
+	}
+
 	return stk->immediate_response;
 }
 
+static void set_get_inkey_duration(struct stk_duration *duration,
+					struct timeval *start_ts)
+{
+	struct timeval end_ts;
+	int interval;
+
+	gettimeofday(&end_ts, NULL);
+
+	interval = (end_ts.tv_usec + 1099999 - start_ts->tv_usec) / 100000;
+	interval += (end_ts.tv_sec - start_ts->tv_sec) * 10;
+	interval -= 10;
+
+	switch (duration->unit) {
+	case STK_DURATION_TYPE_MINUTES:
+		interval = (interval + 59) / 60;
+	case STK_DURATION_TYPE_SECONDS:
+		interval = (interval + 9) / 10;
+	case STK_DURATION_TYPE_SECOND_TENTHS:
+		break;
+	}
+
+	duration->interval = interval;
+}
+
+static void request_confirmation_cb(enum stk_agent_result result,
+					gboolean confirm,
+					void *user_data)
+{
+	struct ofono_stk *stk = user_data;
+	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
+	struct stk_command_get_inkey *cmd = &stk->pending_cmd->get_inkey;
+	struct stk_response rsp;
+
+	stk->respond_on_exit = FALSE;
+
+	switch (result) {
+	case STK_AGENT_RESULT_OK:
+		memset(&rsp, 0, sizeof(rsp));
+
+		rsp.result.type = STK_RESULT_TYPE_SUCCESS;
+		rsp.get_inkey.text.text = confirm ? "" : NULL;
+		rsp.get_inkey.text.yesno = TRUE;
+
+		if (cmd->duration.interval) {
+			rsp.get_inkey.duration.unit = cmd->duration.unit;
+			set_get_inkey_duration(&rsp.get_inkey.duration,
+						&stk->get_inkey_start_ts);
+		}
+
+		if (stk_respond(stk, &rsp, stk_command_cb))
+			stk_command_cb(&error, stk);
+
+		break;
+
+	case STK_AGENT_RESULT_BACK:
+		send_simple_response(stk, STK_RESULT_TYPE_GO_BACK);
+		break;
+
+	case STK_AGENT_RESULT_TIMEOUT:
+		send_simple_response(stk, STK_RESULT_TYPE_NO_RESPONSE);
+		break;
+
+	case STK_AGENT_RESULT_TERMINATE:
+		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
+		break;
+	}
+}
+
+static void request_key_cb(enum stk_agent_result result, char *string,
+				void *user_data)
+{
+	struct ofono_stk *stk = user_data;
+	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
+	struct stk_command_get_inkey *cmd = &stk->pending_cmd->get_inkey;
+	struct stk_response rsp;
+
+	stk->respond_on_exit = FALSE;
+
+	switch (result) {
+	case STK_AGENT_RESULT_OK:
+		memset(&rsp, 0, sizeof(rsp));
+
+		rsp.result.type = STK_RESULT_TYPE_SUCCESS;
+		rsp.get_inkey.text.text = string;
+
+		if (cmd->duration.interval) {
+			rsp.get_inkey.duration.unit = cmd->duration.unit;
+			set_get_inkey_duration(&rsp.get_inkey.duration,
+						&stk->get_inkey_start_ts);
+		}
+
+		if (stk_respond(stk, &rsp, stk_command_cb))
+			stk_command_cb(&error, stk);
+
+		break;
+
+	case STK_AGENT_RESULT_BACK:
+		send_simple_response(stk, STK_RESULT_TYPE_GO_BACK);
+		break;
+
+	case STK_AGENT_RESULT_TIMEOUT:
+		send_simple_response(stk, STK_RESULT_TYPE_NO_RESPONSE);
+		break;
+
+	case STK_AGENT_RESULT_TERMINATE:
+		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
+		break;
+	}
+}
+
+static gboolean handle_command_get_inkey(const struct stk_command *cmd,
+						struct stk_response *rsp,
+						struct ofono_stk *stk)
+{
+	int timeout = stk->timeout * 1000;
+	const struct stk_command_get_inkey *gi = &cmd->get_inkey;
+	uint8_t qualifier = stk->pending_cmd->qualifier;
+	gboolean alphabet = (qualifier & (1 << 0)) != 0;
+	gboolean ucs2 = (qualifier & (1 << 1)) != 0;
+	gboolean yesno = (qualifier & (1 << 2)) != 0;
+	/*
+	 * Note: immediate response and help parameter values are not
+	 * provided by current api.
+	 */
+	uint8_t icon_id = 0;
+	int err;
+
+	if (gi->duration.interval) {
+		timeout = gi->duration.interval;
+		switch (gi->duration.unit) {
+		case STK_DURATION_TYPE_MINUTES:
+			timeout *= 60;
+		case STK_DURATION_TYPE_SECONDS:
+			timeout *= 10;
+		case STK_DURATION_TYPE_SECOND_TENTHS:
+			timeout *= 100;
+		}
+	}
+
+	gettimeofday(&stk->get_inkey_start_ts, NULL);
+
+	if (yesno)
+		err = stk_agent_request_confirmation(stk->current_agent,
+							gi->text, icon_id,
+							request_confirmation_cb,
+							stk, NULL, timeout);
+	else if (alphabet)
+		err = stk_agent_request_key(stk->current_agent, gi->text,
+						icon_id, ucs2, request_key_cb,
+						stk, NULL, timeout);
+	else
+		err = stk_agent_request_digit(stk->current_agent, gi->text,
+						icon_id, request_key_cb,
+						stk, NULL, timeout);
+
+	if (err < 0) {
+		/*
+		 * We most likely got an out of memory error, tell SIM
+		 * to retry
+		 */
+		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		return TRUE;
+	}
+
+	stk->respond_on_exit = TRUE;
+	stk->cancel_cmd = stk_request_cancel;
+
+	return FALSE;
+}
+
+static void request_string_cb(enum stk_agent_result result, char *string,
+				void *user_data)
+{
+	struct ofono_stk *stk = user_data;
+	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
+	uint8_t qualifier = stk->pending_cmd->qualifier;
+	gboolean packed = (qualifier & (1 << 3)) != 0;
+	struct stk_response rsp;
+
+	stk->respond_on_exit = FALSE;
+
+	switch (result) {
+	case STK_AGENT_RESULT_OK:
+		memset(&rsp, 0, sizeof(rsp));
+
+		rsp.result.type = STK_RESULT_TYPE_SUCCESS;
+		rsp.get_input.text.text = string;
+		rsp.get_input.text.packed = packed;
+
+		if (stk_respond(stk, &rsp, stk_command_cb))
+			stk_command_cb(&error, stk);
+
+		break;
+
+	case STK_AGENT_RESULT_BACK:
+		send_simple_response(stk, STK_RESULT_TYPE_GO_BACK);
+		break;
+
+	case STK_AGENT_RESULT_TIMEOUT:
+		send_simple_response(stk, STK_RESULT_TYPE_NO_RESPONSE);
+		break;
+
+	case STK_AGENT_RESULT_TERMINATE:
+		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
+		break;
+	}
+}
+
+static gboolean handle_command_get_input(const struct stk_command *cmd,
+						struct stk_response *rsp,
+						struct ofono_stk *stk)
+{
+	int timeout = stk->timeout * 1000;
+	const struct stk_command_get_input *gi = &cmd->get_input;
+	uint8_t qualifier = stk->pending_cmd->qualifier;
+	gboolean alphabet = (qualifier & (1 << 0)) != 0;
+	gboolean ucs2 = (qualifier & (1 << 1)) != 0;
+	gboolean hidden = (qualifier & (1 << 2)) != 0;
+	uint8_t icon_id = 0;
+	int err;
+
+	if (alphabet)
+		err = stk_agent_request_input(stk->current_agent, gi->text,
+						icon_id, gi->default_text, ucs2,
+						gi->resp_len.min,
+						gi->resp_len.max, hidden,
+						request_string_cb,
+						stk, NULL, timeout);
+	else
+		err = stk_agent_request_digits(stk->current_agent, gi->text,
+						icon_id, gi->default_text,
+						gi->resp_len.min,
+						gi->resp_len.max, hidden,
+						request_string_cb,
+						stk, NULL, timeout);
+
+	if (err < 0) {
+		/*
+		 * We most likely got an out of memory error, tell SIM
+		 * to retry
+		 */
+		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		return TRUE;
+	}
+
+	stk->respond_on_exit = TRUE;
+	stk->cancel_cmd = stk_request_cancel;
+
+	return FALSE;
+}
+
+static void call_setup_connected(struct ofono_call *call, void *data)
+{
+	struct ofono_stk *stk = data;
+	struct stk_response rsp;
+	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
+	static unsigned char facility_rejected_result[] = { 0x9d };
+
+	if (!call) {
+		memset(&rsp, 0, sizeof(rsp));
+
+		rsp.result.type = STK_RESULT_TYPE_NETWORK_UNAVAILABLE;
+		rsp.result.additional_len = sizeof(facility_rejected_result);
+		rsp.result.additional = facility_rejected_result;
+
+		if (stk_respond(stk, &rsp, stk_command_cb))
+			stk_command_cb(&error, stk);
+
+		return;
+	}
+
+	if (call->status == CALL_STATUS_ACTIVE)
+		send_simple_response(stk, STK_RESULT_TYPE_SUCCESS);
+	else
+		send_simple_response(stk, STK_RESULT_TYPE_USER_CANCEL);
+}
+
+static void call_setup_cancel(struct ofono_stk *stk)
+{
+	struct ofono_voicecall *vc;
+	struct ofono_atom *vc_atom;
+
+	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
+						OFONO_ATOM_TYPE_VOICECALL);
+	if (!vc_atom)
+		return;
+
+	vc = __ofono_atom_get_data(vc_atom);
+	if (vc)
+		__ofono_voicecall_dial_cancel(vc);
+}
+
+static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
+				void *user_data)
+{
+	struct ofono_stk *stk = user_data;
+	static struct ofono_error error = { .type = OFONO_ERROR_TYPE_FAILURE };
+	const struct stk_command_setup_call *sc = &stk->pending_cmd->setup_call;
+	uint8_t qualifier = stk->pending_cmd->qualifier;
+	static unsigned char busy_on_call_result[] = { 0x02 };
+	static unsigned char no_cause_result[] = { 0x00 };
+	struct ofono_voicecall *vc = NULL;
+	struct ofono_atom *vc_atom;
+	struct stk_response rsp;
+	int err;
+
+	stk->respond_on_exit = FALSE;
+
+	switch (result) {
+	case STK_AGENT_RESULT_TIMEOUT:
+		confirm = FALSE;
+		/* Fall through */
+
+	case STK_AGENT_RESULT_OK:
+		if (confirm)
+			break;
+
+		send_simple_response(stk, STK_RESULT_TYPE_USER_REJECT);
+		return;
+
+	case STK_AGENT_RESULT_TERMINATE:
+	default:
+		send_simple_response(stk, STK_RESULT_TYPE_USER_TERMINATED);
+		return;
+	}
+
+	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
+						OFONO_ATOM_TYPE_VOICECALL);
+	if (vc_atom)
+		vc = __ofono_atom_get_data(vc_atom);
+
+	if (!vc) {
+		send_simple_response(stk, STK_RESULT_TYPE_NOT_CAPABLE);
+		return;
+	}
+
+	err = __ofono_voicecall_dial(vc, sc->addr.number, sc->addr.ton_npi,
+					sc->alpha_id_call_setup, 0,
+					qualifier >> 1, call_setup_connected,
+					stk);
+	if (err >= 0) {
+		stk->cancel_cmd = call_setup_cancel;
+
+		return;
+	}
+
+	if (err == -EBUSY) {
+		memset(&rsp, 0, sizeof(rsp));
+
+		rsp.result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		rsp.result.additional_len = sizeof(busy_on_call_result);
+		rsp.result.additional = busy_on_call_result;
+
+		if (stk_respond(stk, &rsp, stk_command_cb))
+			stk_command_cb(&error, stk);
+
+		return;
+	}
+
+	if (err == -ENOSYS) {
+		send_simple_response(stk, STK_RESULT_TYPE_NOT_CAPABLE);
+
+		return;
+	}
+
+	memset(&rsp, 0, sizeof(rsp));
+
+	rsp.result.type = STK_RESULT_TYPE_NETWORK_UNAVAILABLE;
+	rsp.result.additional_len = sizeof(no_cause_result);
+	rsp.result.additional = no_cause_result;
+
+	if (stk_respond(stk, &rsp, stk_command_cb))
+		stk_command_cb(&error, stk);
+}
+
+static gboolean handle_command_set_up_call(const struct stk_command *cmd,
+						struct stk_response *rsp,
+						struct ofono_stk *stk)
+{
+	const struct stk_command_setup_call *sc = &cmd->setup_call;
+	uint8_t qualifier = cmd->qualifier;
+	static unsigned char busy_on_call_result[] = { 0x02 };
+	struct ofono_voicecall *vc = NULL;
+	struct ofono_atom *vc_atom;
+	int err;
+
+	if (qualifier > 5) {
+		rsp->result.type = STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD;
+		return TRUE;
+	}
+
+	/*
+	 * Passing called party subaddress and establishing non-speech
+	 * calls are not supported.
+	 */
+	if (sc->ccp.len || sc->subaddr.len) {
+		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
+		return TRUE;
+	}
+
+	vc_atom = __ofono_modem_find_atom(__ofono_atom_get_modem(stk->atom),
+						OFONO_ATOM_TYPE_VOICECALL);
+	if (vc_atom)
+		vc = __ofono_atom_get_data(vc_atom);
+
+	if (!vc) {
+		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
+		return TRUE;
+	}
+
+	if (__ofono_voicecall_is_busy(vc, qualifier >> 1)) {
+		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		rsp->result.additional_len = sizeof(busy_on_call_result);
+		rsp->result.additional = busy_on_call_result;
+		return TRUE;
+	}
+
+	err = stk_agent_confirm_call(stk->current_agent, sc->alpha_id_usr_cfm,
+					0, confirm_call_cb, stk, NULL,
+					stk->timeout * 1000);
+
+	if (err < 0) {
+		/*
+		 * We most likely got an out of memory error, tell SIM
+		 * to retry
+		 */
+		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		return TRUE;
+	}
+
+	stk->respond_on_exit = TRUE;
+	stk->cancel_cmd = stk_request_cancel;
+
+	return FALSE;
+}
+
 static void stk_proactive_command_cancel(struct ofono_stk *stk)
 {
 	if (stk->immediate_response)
@@ -1143,6 +1599,7 @@ static void stk_proactive_command_cancel(struct ofono_stk *stk)
 		stk->cancel_cmd(stk);
 		stk_command_free(stk->pending_cmd);
 		stk->pending_cmd = NULL;
+		stk->cancel_cmd = NULL;
 	}
 }
 
@@ -1217,6 +1674,7 @@ void ofono_stk_proactive_command_notify(struct ofono_stk *stk,
 		case STK_COMMAND_TYPE_GET_INKEY:
 		case STK_COMMAND_TYPE_GET_INPUT:
 		case STK_COMMAND_TYPE_PLAY_TONE:
+		case STK_COMMAND_TYPE_SETUP_CALL:
 			send_simple_response(stk, STK_RESULT_TYPE_NOT_CAPABLE);
 			return;
 
@@ -1268,6 +1726,21 @@ void ofono_stk_proactive_command_notify(struct ofono_stk *stk,
 							&rsp, stk);
 		break;
 
+	case STK_COMMAND_TYPE_GET_INKEY:
+		respond = handle_command_get_inkey(stk->pending_cmd,
+							&rsp, stk);
+		break;
+
+	case STK_COMMAND_TYPE_GET_INPUT:
+		respond = handle_command_get_input(stk->pending_cmd,
+							&rsp, stk);
+		break;
+
+	case STK_COMMAND_TYPE_SETUP_CALL:
+		respond = handle_command_set_up_call(stk->pending_cmd,
+							&rsp, stk);
+		break;
+
 	default:
 		rsp.result.type = STK_RESULT_TYPE_COMMAND_NOT_UNDERSTOOD;
 		break;
@@ -1316,6 +1789,7 @@ static void stk_unregister(struct ofono_atom *atom)
 	if (stk->pending_cmd) {
 		stk_command_free(stk->pending_cmd);
 		stk->pending_cmd = NULL;
+		stk->cancel_cmd = NULL;
 	}
 
 	if (stk->idle_mode_text) {
